package com.taco.swinger;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.util.Locale;
import java.util.ResourceBundle;

import java.awt.Component;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.*;

import org.apache.log4j.Logger;

import com.taco.data.INoReturnMap;
import com.taco.data.NoReturnMapAdapter;
import com.taco.data.INoReturnObservableMap;
import com.taco.data.ObservableMap;
import com.taco.i18n.gui.IFontAdjustedOnLocaleChange;
import com.taco.i18n.gui.FontUtilities;
import com.taco.swinger.text2gui.DispatchingComponentConverter;
import com.taco.text.CollectionConverter;
import com.taco.util.ChainedResourceBundleFactory;
 
/** A panel that lets the user select a font, the color of the font to
 *  use, and the background color.
 */
public class FontChooser extends JPanel 
  implements IFontAdjustedOnLocaleChange, PropertyChangeListener {
    /** Create a new instance that initially selects the Serif font with
     *  the plain style and a point size of 12. The font color is set to
     *  black while the background is set to white. The default locale is
     *  used.
     */
    public FontChooser() {
        this(new Font("Serif", Font.PLAIN, 12), Color.BLACK, Color.WHITE,
             Locale.getDefault());   
    }
      
    public FontChooser(Font initialFont, Color fontColor, 
      Color backgroundColor, Locale locale) {
        super();
        
        if ((initialFont == null) || (fontColor == null) ||
            (backgroundColor == null) || (locale == null)) {
            throw new NullPointerException();
        }    
        
        super.setLocale(locale);
        
        _font = initialFont;
        _fontColor = fontColor;
        _backgroundColor = backgroundColor;        
        initGUI();
    }

    public void setLocale(Locale locale) {
        if (locale == null) {
            throw new NullPointerException();
        }
        
        super.setLocale(locale);        
                
        initGUI();                
    }
    
    /** Initialize the GUI. */    
    public void initGUI() {
        _makeBundle();
        _makeArgMap();
        _configure();
        _addArgMapListeners();

        try {
            FontUtilities.adjustFontsForLocale(this, getLocale());
        } catch (Exception e) {
            _logger.warn("Can't adjust fonts for locale", e);
        }
    }
    
    /** Respond to the argument map changing. */
    public void propertyChange(PropertyChangeEvent event) {
        String propertyName = event.getPropertyName();
        
        Object value = event.getNewValue();
        
        if ("fontName".equals(propertyName)) {
            // Get a new font with the family name, but the same style and
            // size.
            setSelectedFont(new Font((String) value, _font.getStyle(), 
              _font.getSize()));         
        } else if ("fontSize".equals(propertyName)) {
            Font oldFont = _font;
            // Must convert to float so we change the size, not the style 
            setSelectedFont(_font.deriveFont(((Number) value).floatValue()));
        } else if ("fontStyle".equals(propertyName)) {                                            
            setSelectedFont(_font.deriveFont(((Integer) value).intValue()));             
        } else if ("fontColor".equals(propertyName)) {
            if (value == null) {
                throw new NullPointerException();
            }
            setSelectedFontColor((Color) value);
        } else if ("backgroundColor".equals(propertyName)) {
            if (value == null) {
                throw new NullPointerException();
            }            
            setSelectedBackgroundColor((Color) value);
        }                                        
    }
    
    /** Return the currently selected font. 
     *  @beaninfo
     *       bound: true
     */
    public Font getSelectedFont() {
        return _font;   
    }
    
    /** Set the currently selected font. 
     *  @beaninfo
     *       bound: true
     */    
    public void setSelectedFont(Font font) {
        Font oldFont = _font;
        _font = font;
        
        // We rely on ObservableMap to break recursion cycles
        _argMap.putNoReturn("fontName", _font.getFamily(getLocale()), true);
        _argMap.putNoReturn("fontStyle", new Integer(_font.getStyle()), true);
        _argMap.putNoReturn("fontSize", new Integer(_font.getSize()), true);                
        _argMap.putNoReturn("font", _font, true);   
              
        firePropertyChange("selectedFont", oldFont, _font);        
    }
    
    /** Return the currently selected font color. 
     *  @beaninfo
     *       bound: true
     */    
    public Color getSelectedFontColor() {
        return _fontColor;   
    }
    
    /** Set the currently selected font color. 
     *  @beaninfo
     *       bound: true
     */    
    public void setSelectedFontColor(Color fontColor) {
        if (fontColor == null) {
            throw new NullPointerException();
        }
        
        Color oldColor = _fontColor;
        _fontColor = fontColor;
        _argMap.putNoReturn("fontColor", _fontColor, true);
        firePropertyChange("fontColor", oldColor, _fontColor);
    }
        
    /** Return the currently selected background color. 
     *  @beaninfo
     *       bound: true
     */    
    public Color getSelectedBackgroundColor() {
        return _backgroundColor;   
    }        

    /** Set the currently selected background color. 
     *  @beaninfo
     *       bound: true
     */    
    public void setSelectedBackgroundColor(Color backgroundColor) {
        if (backgroundColor == null) {
            throw new NullPointerException();
        }
        
        Color oldColor = _backgroundColor;
        _backgroundColor = backgroundColor;
        _argMap.putNoReturn("backgroundColor", _backgroundColor, true);
        firePropertyChange("backgroundColor", oldColor, _backgroundColor);
    }               

    /** Create a dialog that wraps this instance. If the user selects OK,
     *  the client property "accepted" will be set to <code>Boolean.TRUE</code>.
     *  Otherwise, it will be set to <code>Boolean.FALSE</code>. 
     *  
     *  @param owner The owner of the created dialog, may be <code>null</code>.
     *  @param title The title of the dialog, if <code>null</code>, the default
     *  title will be used.
     *  @param modal True if the returned dialog should be modal.
     *  @return A dialog wrapping this instance, with OK and Cancel buttons.
     */          
    public JDialog createDialog(Component owner, String title, 
      boolean modal) {
        FontChooser fontChooser = new FontChooser();  
        
        final JDialog dialog = new JDialog();
        
        INoReturnMap dialogArgMap = new NoReturnMapAdapter();
        dialogArgMap.putNoReturn("owner", owner);
       
        if (title != null) {           
            dialogArgMap.putNoReturn("dialogTitle", title);
        }
        
        dialogArgMap.putNoReturn("modal", modal ? Boolean.TRUE : Boolean.FALSE);
                
        dialogArgMap.putNoReturn("okListener", new ActionListener() {
          public void actionPerformed(ActionEvent event) {
              FontChooser.this.putClientProperty("accepted", Boolean.TRUE);
              dialog.dispose();
          }            
        });
        
        dialogArgMap.putNoReturn("cancelListener", new ActionListener() {
          public void actionPerformed(ActionEvent event) {
              FontChooser.this.putClientProperty("accepted", Boolean.FALSE);
              dialog.dispose();
          }            
        });
                
        dialogArgMap.putNoReturn("dialogCenterPanel", this);
        
        try {
            DispatchingComponentConverter.DEFAULT_INSTANCE.configureComposite(
              "jdialog", dialog, "dialog", _bundle, dialogArgMap);
        } catch (Exception e) {
            throw new RuntimeException("Can't configure dialog", e);   
        }       
        
        return dialog;        
    }                       
        
    /** A simple test program. */
    public static void main(String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
          public void run() {            
            FontChooser chooser = new FontChooser();
            JDialog dialog = chooser.createDialog(null, null, true);
            dialog.pack();
            dialog.show();
          }
        });  
    }      
    
    protected ResourceBundle _makeBundle() {
        _bundle = ChainedResourceBundleFactory.DEFAULT_INSTANCE.getBundle(
          getClass().getName() + "ResourceBundle", getLocale());
        return _bundle;  
    }
    
    /** Create the argument map and put the desired values into it. This
     *  implementation creates a map, then puts the this instance as a global
     *  with the name "model". Subclasses should use this map to pass
     *  application-specific parameters to the component converter.
     */    
    protected INoReturnObservableMap _makeArgMap() {
        _argMap = new ObservableMap();
        
        _argMap.putNoReturn("fontName", _font.getFamily(), false);
        _argMap.putNoReturn("fontStyle", 
          new Integer(_font.getStyle()), false);
        _argMap.putNoReturn("fontSize", 
          new Integer(_font.getSize()), false);
        _argMap.putNoReturn("fontColor", _fontColor, false);
        _argMap.putNoReturn("backgroundColor", _backgroundColor, false);
        return _argMap;
    }        
    
    /** Configure this panel. */    
    protected void _configure() {
        try {
            CollectionConverter.INSTANCE_COLLECTION_CONVERTER.toObject(
              _bundle, "globals", _argMap, null);
             DispatchingComponentConverter.DEFAULT_INSTANCE.
              configureComposite("jpanel", this, "fontPanel", _bundle, 
               _argMap);
        } catch (Exception e) {
            throw new RuntimeException("Can't configure font chooser", e);
        }
    }
        
    /** Add listeners to the argument map, so that
     *  <code>propertyChange()</code> can respond to changes to values mapped
     *  to observed keys.
     */        
    protected void _addArgMapListeners() {
        _argMap.addPropertyChangeListener("fontName", this);
        _argMap.addPropertyChangeListener("fontSize", this);
        _argMap.addPropertyChangeListener("fontStyle", this);
        _argMap.addPropertyChangeListener("fontColor", this);
        _argMap.addPropertyChangeListener("backgroundColor", this);    
    }    
 
    /** The resource bundle used to create the GUI. */
    protected ResourceBundle _bundle = null;
    
    /** The argument map used to create the GUI and to communicate with it. */
    protected INoReturnObservableMap _argMap = null;     

    /** The currently selected font. */
    protected Font _font = null;
    
    /** The currently selected font color. */
    protected Color _fontColor = null;
    
    /** The currently selected background color. */
    protected Color _backgroundColor = null;        

    private static final Logger _logger = Logger.getLogger(FontChooser.class);
}
