package com.taco.swinger.text;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import java.io.Serializable;
import java.text.ParseException;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import java.util.regex.Matcher;

import java.awt.Frame;
import java.awt.Dialog;
import java.awt.Window; 
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.*;
import javax.swing.event.SwingPropertyChangeSupport;
import javax.swing.text.*;

import org.apache.log4j.Logger;

import com.taco.data.INoReturnMap;
import com.taco.data.NoReturnMapAdapter;
import com.taco.data.INoReturnObservableMap;
import com.taco.data.MapUtilities;
import com.taco.util.ChainedResourceBundleFactory;
import com.taco.text.CPunctuationState;
import com.taco.text.GlobalUtilities;
import com.taco.swinger.SwingInvokeProxyFactory;
import com.taco.swinger.text2gui.*;

/** Implements search and replace on Swing text components.
 *  The status of the search can be monitored by adding a property change
 *  listener for the "status" property. 
 */
public class TextSearcher implements Serializable {
    public TextSearcher(Frame owner, JTextComponent textComponent) {
        this((Window) owner, textComponent, Locale.getDefault());   
    }
    
    public TextSearcher(Frame owner, JTextComponent textComponent, 
      Locale locale) {
        this((Window) owner, textComponent, locale);      
    }          

    public TextSearcher(Dialog owner, JTextComponent textComponent) {
        this((Window) owner, textComponent, Locale.getDefault());   
    }

    public TextSearcher(Dialog owner, JTextComponent textComponent,
      Locale locale) {
        this((Window) owner, textComponent, locale);   
    }        
          
    protected TextSearcher(Window owner, JTextComponent textComponent, 
      Locale locale) {                  
        if (locale == null) {
            throw new NullPointerException();
        }
        _locale = locale;      
        _owner = owner;
        
        _propertyChangeSupport = new SwingPropertyChangeSupport(this);
        
        setTextComponent(textComponent);
        _makeArgMap();                                      
    }

    /** Return the text component that this instance searches. */
    public JTextComponent getTextComponent() {
        return _textComponent;
    }
    
    /** Set the text component that this instance searches.
     *
     *  @throws NullPointerException if <code>textComponent</code> is
     *  <code>null</code>
     */
    public void setTextComponent(JTextComponent textComponent) {
        if (textComponent == null) {
            throw new NullPointerException();
        }        
        _textComponent = textComponent;        
    }

    /** Return the locale that is used for this instance. */
    public Locale getLocale() {
        return _locale;   
    }
    
    /** Set the locale of this instance.
     *
     *  @throws NullPointerException if <code>locale</code> is
     *  <code>null</code>
     *  
     *  @beaninfo
     *       bound: true
     */
    public void setLocale(Locale locale) {
        if (locale == null) {
            throw new NullPointerException();
        }
        
        if (locale.equals(_locale)) {
            return;
        }        
                
        // If the search dialog is currently showing, dispose of it, but
        // redisplay it later.
        boolean mustShow = false;
        if (_searchDialog != null) {                        
            if (_searchDialog.isShowing()) {
                mustShow = true;
                _searchDialog.dispose();   
            }

            if (_searchDialog.isVisible()) {
                _searchDialog.setVisible(false);                
            }          
        }        
        
        Locale oldLocale = locale;
        _locale = locale;
        
        INoReturnObservableMap oldMap = _argMap;
        
        // Remove all globals.
        oldMap.remove(GlobalUtilities.GLOBALS_KEY);
                        
        _argMap = SwingInvokeProxyFactory.makeSwingInvokeProxyMap();   
        
        // Copy the key / values from the old map.
        _argMap.putAll(oldMap, null);
                      
        // Help gc.                      
        oldMap = null;                       
                       
        // Invalidate
        _bundle = null;
        _searchDialog = null;

        if (mustShow) {
            _initGUI();
            _searchDialog.show();
        }
        
        _propertyChangeSupport.firePropertyChange("locale", oldLocale, locale);                
    }
    
    /** Return an action listener that calls <code>find()</code> when
     *  notified.
     */        
    public ActionListener getFindActionListener() {
        if (_findListener == null) {
            _findListener = new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    find();           
                }
            };
        }
        return _findListener;        
    }

    /** Return an action listener that calls <code>findNext()</code> when
     *  notified.
     */                
    public ActionListener getFindNextActionListener() {
        if (_findNextListener == null) {
            _findNextListener = new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    findNext();           
                }
            };
        }        
        return _findNextListener;
    }

    /** Return an action listener that calls <code>findPrevious()</code> when
     *  notified.
     */            
    public ActionListener getFindPreviousActionListener() {
        if (_findPreviousListener == null) {
            _findPreviousListener = new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    findPrevious();           
                }
            };
        }        
        return _findPreviousListener;
    }
    
    /** Return an action listener that calls <code>replaceAndFindNext()</code>
     *  when notified.
     */            
    public ActionListener getReplaceAndFindNextActionListener() {
        if (_replaceAndFindNextListener == null) {
            _replaceAndFindNextListener = new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    replaceAndFindNext();                                                                
                }
            };
        }        
        return _replaceAndFindNextListener;
    }

    /** Return an action listener that calls
     *  <code>replaceAndFindPrevious()</code> when notified.
     */                
    public ActionListener getReplaceAndFindPreviousActionListener() {
        if (_replaceAndFindPreviousListener == null) {
            _replaceAndFindPreviousListener = new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    replaceAndFindPrevious();                                                                
                }
            };
        }        
        return _replaceAndFindPreviousListener;
    }
    
    /** Return an action listener that calls <code>replaceAll()</code> when
     *  notified.
     */                
    public ActionListener getReplaceAllActionListener() {
        if (_replaceAllListener == null) {
            _replaceAllListener = new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    replaceAll();                                            
                }
            };
        }
        return _replaceAllListener;        
    }    

    /** Return the text to search for. Initially, this is the empty string. */    
    public String getSearchText() {
        return (String) MapUtilities.getObjectValue(_argMap, "searchText");
    }
     
    /** Set the text to search for. This may be a regex pattern string. */ 
    public void setSearchText(String text) {
        if (text == null) {
            throw new NullPointerException();
        }        
        _argMap.putNoReturn("searchText", text);    
    }

    /** Return the text to replace the search text with. Initially, this is the
     *  empty string.
     */    
    public String getReplacementText() {
        return (String) MapUtilities.getObjectValue(_argMap, "replacementText");
    }    

    /** Set the text to replace the search text with. If the search text is a
     *  regex, this may contain references to captured groups via the $n
     *  syntax. Also, escape sequences are unescaped.
     */        
    public void setReplacementText(String text) {
        if (text == null) {
            throw new NullPointerException();
        }                
        _argMap.putNoReturn("replacementText", text);
    }

    /** Return <code>true</code> iff comparisons should be made
     *  case-insensitively. Initially, this is <code>false</code>.
     */    
    public boolean getIgnoreCase() {
        return MapUtilities.getBooleanValue(_argMap, "ignoreCase");
    }
    
    /** If set to <code>true</code>, comparisons will be made
     *  case-insensitively.
     */
    public void setIgnoreCase(boolean flag) {
        _argMap.putNoReturn("ignoreCase", flag ? Boolean.TRUE : Boolean.FALSE);        
    }

    /** Return <code>true</code> iff the search string is to be interpreted as
     *  a regex pattern. Initially, this is <code>false</code>.
     */        
    public boolean getIsRegex() {
        return MapUtilities.getBooleanValue(_argMap, "isRegex");
    }        
    
    /** If set to <code>true</code>, the search string will be interpreted as a
     *  regex pattern.
     */           
    public void setIsRegex(boolean flag) {    
        _argMap.putNoReturn("isRegex", flag ? Boolean.TRUE : Boolean.FALSE);        
    }

    /** Return <code>true</code> iff regex patterns are compiled with the
     *  <code>Pattern.UNICODE_CASE</code> option. Initially, this is
     *  <code>false</code>.
     */
    public boolean getUnicodeCase() {
        return MapUtilities.getBooleanValue(_argMap, "unicodeCase");
    }
    
    /** If set to <code>true</code> and the search string is interpreted as a
     *  regex pattern, the pattern will be compiled with the
     *  <code>Pattern.UNICODE_CASE</code> option.
     */      
    public void setUnicodeCase(boolean flag) {
        _argMap.putNoReturn("unicodeCase", 
          flag ? Boolean.TRUE : Boolean.FALSE);
    }

    /** Return <code>true</code> iff regex patterns are compiled with the
     *  <code>Pattern.DOTALL</code> option. Initially, this is
     *  <code>false</code>.
     */    
    public boolean getDotAll() {
        return MapUtilities.getBooleanValue(_argMap, "dotAll");
    }

    /** If set to <code>true</code> and the search string is interpreted as a
     *  regex pattern, the pattern will be compiled with the
     *  <code>Pattern.DOTALL</code> option.
     */      
    public void setDotAll(boolean flag) {
        _argMap.putNoReturn("dotAll", flag ? Boolean.TRUE : Boolean.FALSE);
    }
    
    /** Return <code>true</code> iff regex patterns are compiled with the
     *  <code>Pattern.MULTILINE</code> option. Initially, this is
     *  <code>false</code>.
     */    
    public boolean getMultiLine() {
        return MapUtilities.getBooleanValue(_argMap, "multiLine");
    }

    /** If set to <code>true</code> and the search string is interpreted as a
     *  regex pattern, the pattern will be compiled with the
     *  <code>Pattern.MULTILINE</code> option.
     */      
    public void setMultiLine(boolean flag) {
        _argMap.putNoReturn("multiLine", flag ? Boolean.TRUE : Boolean.FALSE);
    }

    /** Return <code>true</code> iff regex patterns are compiled with the
     *  <code>Pattern.COMMENTS</code> option. Initially, this is
     *  <code>false</code>.
     */     
    public boolean getAllowWhiteSpace() {
        return MapUtilities.getBooleanValue(_argMap, "allowWhiteSpace");
    }        

    /** If set to <code>true</code> and the search string is interpreted as a
     *  regex pattern, the pattern will be compiled with the
     *  <code>Pattern.COMMENTS</code> option.
     */      
    public void setAllowWhiteSpace(boolean flag) {
        _argMap.putNoReturn("allowWhiteSpace", 
          flag ? Boolean.TRUE : Boolean.FALSE);
    }
   
    /** Return <code>true</code> iff escape sequences are to be unescaped in
     *  the search text. Initially, this is <code>false</code>.
     */
    public boolean getUnescapeSearchText() {
        return MapUtilities.getBooleanValue(_argMap, "unescapeSearchText");
    }

    /** If set to <code>true</code>, the escaped sequences in the search text
     *  will be unescaped. In combination with <code>setIsRegex()</code>, this
     *  allows regex strings to be entered as written in Java code. Initially,
     *  this is <code>false</code>.
     */    
    public void setUnescapeSearchText(boolean flag) {
        _argMap.putNoReturn("unescapeSearchText", 
          flag ? Boolean.TRUE : Boolean.FALSE);
    }

    /** Return <code>true</code> iff escape sequences are to be unescaped in
     *  the replacement text. Initially, this is <code>false</code>.
     */    
    public boolean getUnescapeReplacementText() {
        return MapUtilities.getBooleanValue(_argMap, "unescapeReplacementText");
    }

    /** If set to <code>true</code>, the escaped sequences in the replacement
     *  text will be unescaped.
     */        
    public void setUnescapeReplacementText(boolean flag) {
        _argMap.putNoReturn("unescapeReplacementText", 
          flag ? Boolean.TRUE : Boolean.FALSE);
    }
        
    /** Return true iff searches are to operate in the current selection only.
     *  Initially, this is <code>false</code>.
     */
    public boolean getSearchSelectionOnly() {
        return MapUtilities.getBooleanValue(_argMap, "searchSelection");
    }

    /** If set to <code>true</code>, searches will operate in the current
     *  selection only.
     */    
    public void setSearchSelectionOnly(boolean flag) {        
        _argMap.putNoReturn("searchSelection", 
          flag ? Boolean.TRUE : Boolean.FALSE);
        _argMap.putNoReturn("searchEntire",    
          flag ? Boolean.FALSE : Boolean.TRUE);
    }
    
    /** Return <code>true</code> iff the search dialog should be kept open
     *  after one of its search buttons are pressed. Initially, this is
     *  <code>true</code>.
     */            
    public boolean getKeepDialog() {
        return MapUtilities.getBooleanValue(_argMap, "keepDialog");
    }
    
    /** If set to <code>true</code>, the search dialog will be kept open
     *  after one of its search buttons are pressed.
     */    
    public void setKeepDialog(boolean flag) {
        _argMap.putNoReturn("keepDialog", flag ? Boolean.TRUE : Boolean.FALSE);        
    }
    
    /** Return <code>true</code> if searches should automatically wrap from the
     *  end to beginning or vice versa after reaching a boundary. Initially,
     *  this is <code>false</code>.
     */   
    public boolean getAutoWrap() {
        return MapUtilities.getBooleanValue(_argMap, "autoWrap");
    }
       
    /** If set to <code>true</code>, searches will automatically wrap from the
     *  end to beginning or vice versa after reaching a boundary.
     */       
    public void setAutoWrap(boolean flag) {    
        _argMap.putNoReturn("autoWrap", flag ? Boolean.TRUE : Boolean.FALSE);        
    }        

    /** Show the find and replace dialog. */            
    public void find() {
        if (_searchDialog == null) {
            _initGUI();
        }                 
        
        updateSelection();                
        _searchDialog.show();        
    }
    
    /** Find the next occurance of the search text. If the search text is
     *  empty, show the search dialog and return. If the search text found,
     *  select the occurance.
     */
    public void findNext() {
        _findNext(COMMAND_FIND, null);
    }

    /** Find the previous occurance of the search text. If the search text is
     *  empty, show the search dialog and return. If the search text found,
     *  select the occurance.
     */
    public void findPrevious() {
        _findPrevious(COMMAND_FIND, null);
    }
    
    /** Replace the current selection and find the next occurance. */
    public void replaceAndFindNext() {
        _replace(COMMAND_REPLACE_AND_FIND, DIRECTION_FORWARDS);   
    }
    
    /** Replace the current selection and find the previous occurance. */
    public void replaceAndFindPrevious() {
        _replace(COMMAND_REPLACE_AND_FIND, DIRECTION_BACKWARDS);   
    }
    
    /** Replace all occurrances of the search text with the replacement 
     *  text. 
     */
    public void replaceAll() {
        _replace(COMMAND_REPLACE_ALL, DIRECTION_FORWARDS);   
    }
        
    /** Replace the selected portion with the current replacement text. If
     *  there is no selection, return <code>null</code>. Otherwise, do the
     *  replacement and return an array containing the position of the
     *  beginning of the replacement text and the position of the end of the
     *  replacement text. This method does not set the status of this instance.
     */
    public int[] replaceSelection() {
        Document document = _textComponent.getDocument();
        Caret caret = _textComponent.getCaret();
        
        int dot = caret.getDot();
        int mark = caret.getMark();
        
        // Return null if there was no selection.
        if (dot == mark) {
            return null;
        }
        
        // Make it so that the dot is always before the mark.
        if (dot > mark) {
            int temp = dot;
            dot = mark;
            mark = temp;
        }
            
        int selectionLength = mark - dot;            
                      
        String replacementText = getReplacementText();                      
                      
        // Substitue backreferences, and unescape.                       
        replacementText = _interpolate(replacementText, true);                                                                               
                     
        try {
            // Remove the selected text.             
            document.remove(dot, selectionLength);
        
            // Insert the replacement text.
            document.insertString(dot, replacementText, null);                        
        } catch (BadLocationException e) {
            throw new RuntimeException("couldn't replace selection between " +
             dot + " and " + mark, e);
        }        
        
        int replacementLength = replacementText.length(); 
        int[] retval = new int[] { dot, dot + replacementLength };
        
        // Now that we have changed the document, update the selection indices
        // so that the same region is selected.
        if (_selectionEnd >= 0) {                       
            _selectionEnd += (replacementLength - selectionLength); 
        }        
        
        return retval;        
    }

    /** Determine if a selection exists, and if so, update the search dialog
     *  appropriately.    
     */
    public void updateSelection() {                                
        Caret caret = _textComponent.getCaret();
        int dot = caret.getDot();
        int mark = caret.getMark();
                                        
        if (dot != mark) {
            // There is a selection, so have the search dialog box default to
            // searching in the selection only.
            _argMap.putNoReturn("searchSelectionEnabled", Boolean.TRUE, true);
            setSearchSelectionOnly(true);
              
            if (dot < mark) { 
                _selectionStart = dot;
                _selectionEnd = mark;
            } else {
                _selectionEnd = dot;
                _selectionStart = mark;
            }                                                                       
        } else {
            // There is no selection, so have the search dialog box default to
            // searching the entire document.
            _argMap.putNoReturn("searchSelectionEnabled", Boolean.TRUE, true);
            _argMap.putNoReturn("searchEntire", Boolean.TRUE, true);
            _argMap.putNoReturn("searchSelectionEnabled", Boolean.FALSE, true);
                        
            _selectionStart = -1;
            _selectionEnd = -1;
        }                                    
    }
    
    /** Return the search status. */
    public int getStatus() {
        return _status;
    }
    
    /** Return the extra info associated with the status. */
    public Object getExtraInfo() {
        return _extraInfo;   
    }

    public boolean hasListeners(String propertyName) {
        if (propertyName == null) {
            throw new NullPointerException();
        }
        return _propertyChangeSupport.hasListeners(propertyName);   
    }
            
    public PropertyChangeListener[] getPropertyChangeListeners() {
        return _propertyChangeSupport.getPropertyChangeListeners();
    }
    
    public PropertyChangeListener[] getPropertyChangeListeners(
      String propertyName) {
        if (propertyName == null) {
            throw new NullPointerException();
        }          
        return _propertyChangeSupport.getPropertyChangeListeners(propertyName);
    }
    
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        if (listener == null) {
            throw new NullPointerException();   
        }
        _propertyChangeSupport.addPropertyChangeListener(listener);
    }
                 
    public void addPropertyChangeListener(String propertyName, 
      PropertyChangeListener listener) {
        if ((propertyName == null) || (listener == null)) {
            throw new NullPointerException();   
        }
        _propertyChangeSupport.addPropertyChangeListener(propertyName, 
          listener);
    }
    
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        if (listener == null) {
            throw new NullPointerException();   
        }
        _propertyChangeSupport.removePropertyChangeListener(listener);           
    }
 
    public void removePropertyChangeListener(String propertyName, 
      PropertyChangeListener listener) {  
        if ((propertyName == null) || (listener == null)) {
            throw new NullPointerException();   
        }
        _propertyChangeSupport.removePropertyChangeListener(propertyName, 
          listener);
    }      
                
    protected void setStatus(int status, Object extraInfo) {
        if (_logger.isDebugEnabled()) {
            _logger.debug("status = " + status);
        }
        int oldStatus = _status;
        
        _status = status;
        _extraInfo = extraInfo;
                
        _propertyChangeSupport.firePropertyChange("status", 
          new Integer(oldStatus), new Integer(_status));    
    }
    
    /** Create the GUI components of this instance. */
    protected void _initGUI() {
        _makeBundle();        
        _makeDialog();                     
    }        
   
    /** Create the resource bundle used for the GUI components of this
     *  instance.
     */
    protected ResourceBundle _makeBundle() {
        _bundle = ChainedResourceBundleFactory.DEFAULT_INSTANCE.getBundle(
          "com.taco.swinger.text.SearchResourceBundle", _locale);
        return _bundle;  
    } 
    
    /** Create the argument map used to store settings and to control the
     *  GUI. Put default settings into the map.
     */
    protected INoReturnObservableMap _makeArgMap() {
        _argMap = SwingInvokeProxyFactory.makeSwingInvokeProxyMap();   
        
        setSearchText("");
        setIgnoreCase(false);
        setIsRegex(false);
        setUnescapeSearchText(false);
        
        setReplacementText("");
        setUnescapeReplacementText(false);
        
        setAutoWrap(false);
        setKeepDialog(true);
        setSearchSelectionOnly(false);
                                
        _argMap.putNoReturn("searchSelectionEnabled", Boolean.FALSE, null);                
        _argMap.putNoReturn("owner", _owner, null);        
        _argMap.putNoReturn("findListener", getFindActionListener(), null);
        _argMap.putNoReturn("findNextListener", getFindNextActionListener(), 
          null);
        _argMap.putNoReturn("findPreviousListener", 
          getFindPreviousActionListener(), null);
        _argMap.putNoReturn("replaceAndFindNextListener", 
          getReplaceAndFindNextActionListener(), null);
        _argMap.putNoReturn("replaceAndFindPreviousListener", 
          getReplaceAndFindPreviousActionListener(), null);          
        _argMap.putNoReturn("replaceAllListener", 
          getReplaceAllActionListener(), null);                    
             
        return _argMap;
    }        
    
    /** Create the search dialog. */
    protected JDialog _makeDialog() {
        try {
            _searchDialog = (JDialog) 
              DispatchingComponentConverter.DEFAULT_INSTANCE.toComponent(
              _bundle, "searchDialog", _argMap);
        } catch (Exception e) {
            throw new RuntimeException(e);   
        }
        _searchDialog.pack();
        return _searchDialog;
    }   
    
    protected void _findNext(int command, Pattern pattern) {
        _savedGroups = null;
                        
        // If we haven't searched for anything yet, get the search options 
        // from the user.
        int searchTextLength = getSearchText().length();        
        
        if (searchTextLength == 0) {
            if (_searchDialog == null) {
                _initGUI();
            }
            _searchDialog.show();            
            return;
        }

        boolean searchSelection = getSearchSelectionOnly();            
        boolean keepDialog = getKeepDialog();                                     
        Caret caret = _textComponent.getCaret();        
        
        Document document = _textComponent.getDocument(); 
        CharSequence charSequence = new DocumentCharSequence(document);
                                
        // Depending on if our search scope is the selection or entire document,
        // figure out the index of the first character to look at and a 
        // CharSequence corresponding to all the characters to look at.                
        int startIndex = -1;
                    
        try {                        
            if (searchSelection) {                        
                // If the scope was the selection, but there was no selection, fail.
                
                if ((_selectionStart < 0) || (_selectionEnd < 0)) {
                    setStatus(STATUS_NOT_FOUND, EMPTY_SELECTION);
                    return;
                }
                
                // Get the maximum index if there's a selection.                      
                startIndex = Math.max(_selectionStart,
                  Math.max(caret.getDot(), caret.getMark()));
                
                // If the maximum index is outside the selection, just 
                // start from the beginning of the selection.
                // Include the boundary so that just after making the 
                // selection, we start from the beginning.
                if (startIndex >= _selectionEnd) {
                    startIndex = _selectionStart;
                }                
                
                charSequence = charSequence.subSequence(startIndex, 
                  _selectionEnd);            
            } else {
                // Use the part of the document after the cursor.                        
                
                startIndex = caret.getDot();
                                                                                                                            
                // Move the charSequence to the position of the caret.
                charSequence = charSequence.subSequence(startIndex, 
                 charSequence.length());
            }                   
    
            int[] matchIndices = _findNext(charSequence, pattern);
                                                 
            if (matchIndices != null) {
                int[] indexArray = new int[2];
                indexArray[0] = startIndex + matchIndices[0];
                indexArray[1] = startIndex + matchIndices[1]; 
                                
                // Highlight the matched expression, leaving the caret at the end.                         
                caret.setDot(indexArray[0]);                       
                caret.moveDot(indexArray[1]);
                
                setStatus(STATUS_FOUND, indexArray);
                return;
            } 
            
            if (!searchSelection && (command != COMMAND_REPLACE_ALL) &&
                (_status != STATUS_HIT_BOUNDARY)) {                                                    
                // If we reach the end and we haven't wrapped before, wrap 
                // to the beginning automatically or give the user a chance
                // to request the same thing.                                         
                if (getAutoWrap() || _askToWrap(DIRECTION_FORWARDS)) {
                    setStatus(STATUS_HIT_BOUNDARY, Boolean.FALSE);
                    caret.setDot(0);                        
                    _findNext(command, pattern);            
                    return;                       
                } 
            }
                                            
            setStatus(STATUS_NOT_FOUND, null);    
        } finally {
            if (!keepDialog && (_searchDialog != null)) {
                _searchDialog.hide();   
            }            
        }                                            
    }
        
    protected void _findPrevious(int command, Pattern pattern) {
        _savedGroups = null;
                 
        int searchTextLength = getSearchText().length();         
        
        // If we haven't searched for anything yet, call find() to get the 
        // search options from the user.
        if (searchTextLength == 0) {
            if (_searchDialog == null) {
                _initGUI();
            }            
            _searchDialog.show();
            return;            
        }        
        
        boolean searchSelection = getSearchSelectionOnly();           
        boolean keepDialog = getKeepDialog(); 
                    
        Document document = _textComponent.getDocument(); 
        CharSequence charSequence = new DocumentCharSequence(document);
 
        Caret caret = _textComponent.getCaret();
        
        // Depending on if our search scope is the selection or entire document,
        // figure out the index of the first character to look at and a 
        // CharSequence corresponding to all the characters to look at.                
        int startIndex;

        try {                
            if (searchSelection) {
                // If the scope was the selection, but there was no selection, fail.                                            
                if ((_selectionStart < 0) || (_selectionEnd < 0)) {
                    setStatus(STATUS_NOT_FOUND, EMPTY_SELECTION);                
                    return;
                }
                
                startIndex = _selectionStart;
                
                // Get the minimum index if there's a selection.                      
                int endIndex = Math.min(_selectionEnd,
                  Math.min(caret.getDot(), caret.getMark()));
                
                // If the minimum index is outside the selection, just 
                // start from the end of the selection.
                // Include the boundary so that just after making the 
                // selection, we search the entire selection.
                if (endIndex <= _selectionStart) {
                    endIndex = _selectionEnd;
                }
                
                charSequence = charSequence.subSequence(_selectionStart, 
                  endIndex);            
            } else {
                // Use the part of the document before the cursor.           
                
                startIndex = 0;
                                                                                                                            
                // Truncate the charSequence at the position of the caret.
                charSequence = charSequence.subSequence(0, caret.getDot());
            }                   
    
            int[] matchIndices = _findPrevious(charSequence, pattern);                                                                         
                                                                                                                                                      
            if (matchIndices != null) {
                // Highlight the matched expression, leaving the caret at the 
                // beginning.
                int[] indexArray = new int[2];
                indexArray[0] = startIndex + matchIndices[0];                  
                indexArray[1] = startIndex + matchIndices[1];
                
                caret.setDot(indexArray[1]);
                caret.moveDot(indexArray[0]);
    
                setStatus(STATUS_FOUND, indexArray);
                return;                        
            } 
    
            if (!searchSelection && (command != COMMAND_REPLACE_ALL) &&
                (_status != STATUS_HIT_BOUNDARY)) {                     
                    
                // If we reach the beginning and we haven't wrapped before,
                // wrap to the end automatically or give the user a chance to
                // request the same thing.                                       
                if (getAutoWrap() || _askToWrap(DIRECTION_BACKWARDS)) {
                    setStatus(STATUS_HIT_BOUNDARY, Boolean.TRUE);
                    caret.setDot(document.getLength() - 1);                        
                    _findPrevious(command, pattern);            
                    return;                       
                } 
            }
            
            setStatus(STATUS_NOT_FOUND, null);
        } finally {
            if (!keepDialog && (_searchDialog != null)) {
                _searchDialog.hide();   
            }                                        
        }                                
    }
        
    /** Find the next occurance of the search text in
     *  <code>charSequence</code>. If successful, return a 2 element int array,
     *  with the first element being the start index of the matched sequence,
     *  and the second element being the end index (not inclusive) . If not
     *  match was found, return <code>null</code>.
     */
    protected int[] _findNext(CharSequence charSequence, Pattern pattern) {
        // Unescape if necessary.                    
        String searchText = _interpolate(getSearchText(), false);
        
        boolean ignoreCase = getIgnoreCase();
        
        if (getIsRegex()) {
            // Compute the pattern now if not already computed.
            if (pattern == null) {
                try {
                    pattern = _computePattern();
                } catch (PatternSyntaxException pse) {
                    setStatus(STATUS_PATTERN_ERROR, pse);
                    return null;
                }
            }
                                                                                                           
            Matcher matcher = pattern.matcher(charSequence);
            
            if (matcher.find()) {
                _saveGroups(matcher);
                return new int[] { matcher.start(), matcher.end() };
            } else {
                _savedGroups = null;
            }                        
        } else {
            _savedGroups = null;
                            
            int searchTextLength = searchText.length();                                         
                             
            for (int i = 0; i <= charSequence.length() - searchTextLength; 
                 i++) {
                boolean matches = true;
                
                // Look for a character that doesn't match, and when that happens,
                // set matches to false.
                
                if (ignoreCase) {
                    // compare the upper case versions to ignore the case
                    for (int j = 0; j < searchTextLength; j++) {                    
                        if (Character.toUpperCase(charSequence.charAt(i + j)) != 
                            Character.toUpperCase(searchText.charAt(j))) {
                            matches = false;
                            break;
                        }
                    }
                } else {                    
                    for (int j = 0; j < searchTextLength; j++) {                    
                        if (charSequence.charAt(i + j) != searchText.charAt(j)) {
                            matches = false;
                            break;
                        }
                    }                                        
                }
                
                // If we still have a match, we are done.
                if (matches) {                    
                    return new int[] { i, i + searchTextLength };  
                }
            }                        
        }   
        
        // No match was found.
        return null;             
    }

    /** Find the last occurance of the search text in
     *  <code>charSequence</code>. If successful, return a 2 element int array,
     *  with the first element being the start index of the matched sequence,
     *  and the second element being the end index (not inclusive) . If not
     *  match was found, return <code>null</code>.
     */    
    protected int[] _findPrevious(CharSequence charSequence, Pattern pattern) {
        // Unescape if necessary.
        String searchText = _interpolate(getSearchText(), false);

        boolean ignoreCase = getIgnoreCase();
                                                                                  
        int sequenceLength = charSequence.length();            
                                             
        if (getIsRegex()) {
            // Compute the pattern now if not already computed.
            if (pattern == null) {
                try {
                    pattern = _computePattern();
                } catch (PatternSyntaxException pse) {
                    setStatus(STATUS_PATTERN_ERROR, pse);
                    return null;
                }
            }
                                                                 
            Matcher matcher = null;
                
            // Try to match the pattern using the smallest subsequence of the
            // sequence we are looking at. Increase the size of the subsequence
            // to n as we fail matching the last n - 1 characters. 
            int index;             
            
            for (index = sequenceLength - 1; index >= 0; index--) {
                // Try to match the subsequence starting from index.             
                CharSequence subSequence = 
                 charSequence.subSequence(index, sequenceLength);
                matcher = pattern.matcher(subSequence);                                 
                
                if (matcher.lookingAt()) {
                    _saveGroups(matcher);
                    return new int[] { index, index + matcher.end() }; 
                } else {
                    _savedGroups = null;
                }
            }            
        } else { 
            _savedGroups = null;
            
            int searchTextLength = searchText.length();
                                             
            for (int i = sequenceLength - searchTextLength; i >= 0; i--) {
                boolean matches = true;
                
                // Look for a character that doesn't match, and when that happens,
                // set matches to false.
                
                if (ignoreCase) {
                    // compare the upper case versions to ignore the case
                    for (int j = 0; j < searchTextLength; j++) {                    
                        if (Character.toUpperCase(charSequence.charAt(i + j)) != 
                            Character.toUpperCase(searchText.charAt(j))) {
                            matches = false;
                            break;
                        }
                    }
                } else {                    
                    for (int j = 0; j < searchTextLength; j++) {                    
                        if (charSequence.charAt(i + j) != searchText.charAt(j)) {
                            matches = false;
                            break;
                        }
                    }                                        
                }
                
                // If we still have a match, we are done.
                if (matches) {
                    return new int[] { i, i + searchTextLength };
                }
            }
        }
        
        // None of the subsequences matched.            
        return null;                
    }                     

    /** Return the number of replacements made. */        
    protected int _replace(int command, int direction) {
        boolean searchSelection = getSearchSelectionOnly();                   
                        
        boolean keepGoingIfFound = false;
        Caret caret = _textComponent.getCaret();        
        
        Pattern pattern = null;
        
        if (command == COMMAND_REPLACE_ALL) {
            // Set the cursor to the beginning of the selection, or the
            // beginning of the document, depending on the scope of the
            // search request.                
            if (searchSelection) {
                if ((_selectionStart < 0) || (_selectionEnd < 0)) {
                    setStatus(STATUS_NOT_FOUND, EMPTY_SELECTION);
                    return 0;
                }
                
                caret.setDot(_selectionStart);
            } else {
                caret.setDot(0);
            }

            // Pre-compute the pattern here, so it can be reused.
            if (getIsRegex()) {
                try {
                    pattern = _computePattern();
                } catch (PatternSyntaxException pse) {
                    setStatus(STATUS_PATTERN_ERROR, pse);
                    return 0;
                }
            }

            // Do an initial search here because replacement is done first
            // in the loop.
            _findNext(command, pattern);
    
            // If no occurances were found, give up now.            
            if (_status == STATUS_NOT_FOUND) {
                return 0;
            }        
                                  
            keepGoingIfFound = true;
            direction = DIRECTION_FORWARDS;
        } else if (command == COMMAND_REPLACE_AND_FIND) {
            keepGoingIfFound = false;                                    
        } else {
            throw new IllegalArgumentException("Unknown command: " +
              command);
        }
        

        int numReplacements = 0;
                                        
        do {            
            // Try to replace the selection with the replacement text. If
            // there was no selection, fail.                     
            int[] indexArray = replaceSelection(); 
                                        
            if (indexArray == null) {
                return 0;
            } 
            
            setStatus(STATUS_REPLACED_TEXT, indexArray);
    
            int beginIndex = -1; 
    
            // Now position the cursor after or before the replacement
            // text, so searching in the same direction won't look over the
            // replacement text. If we are going forwards, this means we
            // need to set the caret to after the replacement text. If we
            // are going backwards, this means we need to set the caret to
            // before the replacement text.
            if (direction ==  DIRECTION_FORWARDS) {                
                beginIndex = indexArray[1];
            } else {
                beginIndex = indexArray[0];   
            }
            
            caret.setDot(beginIndex);
            
            numReplacements++;
                     
            if (direction == DIRECTION_FORWARDS) {
                _findNext(command, pattern);
            } else {
                _findPrevious(command, pattern);
            }                        
        } while (keepGoingIfFound && (_status != STATUS_NOT_FOUND));  

        setStatus(STATUS_REPLACE_ALL_FINISHED, new Integer(numReplacements));
                        
        return numReplacements;
    } 
                
    /** Ask the user if he wants to wrap the search around once the beginning
     *  or end of the document is reached.
     */
    protected boolean _askToWrap(int direction) {
        INoReturnMap argMap = new NoReturnMapAdapter();
        
        argMap.putNoReturn("wrap", Boolean.FALSE);
        argMap.putNoReturn("backwards", 
         ((direction == DIRECTION_BACKWARDS) ? Boolean.TRUE : Boolean.FALSE));
        argMap.putNoReturn("owner", _owner); 
        
        JDialog dialog = null;
        try {
            dialog = (JDialog) 
              DispatchingComponentConverter.DEFAULT_INSTANCE.toComponent(
              _bundle, "wrapDialog", argMap);
        } catch (Exception e) {
            throw new RuntimeException("Couldn't create wrap dialog", e);   
        }
        
        dialog.pack();
        dialog.show();
        
        return Boolean.TRUE.equals(argMap.get("wrap"));        
    }
            
    /** If necessary, unescape the text, depending on the options and whether
     *  or not the text is the replacement text. Also, if the text is the
     *  replacement text and the search text is a regex pattern, replace
     *  occurances of $1, $2, etc. with what was matched in the corresponding
     *  group.
     */            
    protected String _interpolate(CharSequence text, 
      boolean isReplacementText) {
        if (text == null) {
            throw new NullPointerException();
        }
          
        StringBuffer sb = new StringBuffer();
        
        boolean unescape = false;
        
        if (isReplacementText) {
            unescape = getUnescapeReplacementText();
        } else {
            unescape = getUnescapeSearchText();                                    
        }
                
        boolean backReference = isReplacementText && getIsRegex();
        
        // Shortcut: if we're not going to be doing anything special below,
        // just return the text unchanged.
        if (!unescape && !backReference) {
            return text.toString();                   
        }        
        
        int groups = 0;
        if (_savedGroups != null) {
            groups = _savedGroups.length;
        }
                                        
        int i = 0;        
        int length = text.length(); 
        while (i < length) {
            char c = text.charAt(i);
            
            if (backReference && (c == '$')) {                
                int groupIndex = 0;
                i++; 

                boolean foundDigit = false;

                DigitLoop:                
                for (; i < length; i++) {
                    char d = text.charAt(i);
                    
                    if ((d >= '0') && (d <= '9')) {
                        foundDigit = true;
                        groupIndex = groupIndex * 10 + (d - '0');
                    } else {
                        break DigitLoop;   
                    }
                }                        
                
                if (foundDigit) {
                    if (groupIndex < groups) {
                        sb.append(_savedGroups[groupIndex]);   
                    } else {
                        // Invalid group index. Just append the raw characters
                        sb.append('$').append("" + groupIndex);                        
                    }
                } else {
                    // No digits after the $, just treat it as an ordinary $                   
                    sb.append('$');
                }                    
            } else if (c == '\\') {
                i++;
                if (i == length) {
                    // If a backslash is the last character, treat it as a
                    // real backslash.
                    sb.append('\\');                    
                } else if (unescape) {                                                                     
                    // Handle escape sequences.
                    CPunctuationState state = new CPunctuationState();

                    try {                                        
                        state.update('\\', i);
                    } catch (ParseException pe) {
                        assert false;
                    }
                
                    char d = Character.MAX_VALUE;
                    
                    EscapedCharsLoop:
                    for (; (i < length) && state.escaped; i++) {
                        try {
                            d = state.update(text.charAt(i), i);

                            if (d != Character.MAX_VALUE) {
                                sb.append(d);   
                            }
                        } catch (ParseException pe2) {
                            break EscapedCharsLoop;   
                        }                        
                    }                    
                } else if (backReference) { 
                    char d = text.charAt(i);
                    
                    // Unescape only $ and '\'
                    if ((d == '$') || (d == '\\')) {
                        sb.append(d);
                        i++;
                    } else {
                        sb.append('\\');   
                    }
                } else {
                    // Just treat as an ordinary '\'
                    sb.append('\\');   
                }
            } else {
                sb.append(c);
                i++;
            }
        }
        
        return sb.toString();                            
    } 
    
    /** Compute the regex pattern based on the current settings. */
    protected Pattern _computePattern() throws PatternSyntaxException {        
        int options = 0;
            
        if (getIgnoreCase()) {
            options |= Pattern.CASE_INSENSITIVE;
        }
        
        if (getUnicodeCase()) {
            options |= Pattern.UNICODE_CASE;   
        }
        
        if (getDotAll()) {
            options |= Pattern.DOTALL;
        }
        
        if (getAllowWhiteSpace()) {
            options |= Pattern.COMMENTS;   
        }
        
        if (getMultiLine()) {
            options |= Pattern.MULTILINE;                
        }   
        
        String searchText = (String) _argMap.get("searchText");
        searchText = _interpolate(searchText, false);
                
        return Pattern.compile(searchText, options);
    }
    
    /** Save the matched groups in <code>_savedGroups</code>. */
    protected void _saveGroups(Matcher matcher) {
        int groups = matcher.groupCount() + 1;
        
        _savedGroups = new String[groups];

        for (int i = 0; i < groups; i++) {
            _savedGroups[i] = matcher.group(i);
        }        
    }
                            
    protected Locale _locale = null;
    
    protected Window _owner = null;
    
    /** The text component that this instance searches. */
    protected JTextComponent _textComponent = null;
    
    /** The resource bundle used to create the GUI. */
    protected ResourceBundle _bundle = null;
    
    /** The argument map used to create the GUI and to communicate with it. */
    protected INoReturnObservableMap _argMap = null;        
 
    protected JDialog _searchDialog = null;
    
    /** The index of the beginning of the selection, at the time 
     *  <code>updateSelection()</code> was last called. If -1, there is no
     *  selection.
     */
    protected int _selectionStart = -1;
    
    /** The index of the end of the selection, at the time
     *  <code>updateSelection()</code> was last called. If -1, there is no
     *  selection.
     */    
    protected int _selectionEnd = -1;
    
    /** The current status. */
    protected int _status = STATUS_INITIAL;
    
    /** The extra information associated with the current status. */
    protected Object _extraInfo = null;
 
    /** The cached regex pattern. */
    protected Pattern _savedPattern = null;
 
    /** If the last match was by a regex, this array stores the groups that
     *  were matched. Otherwise, it is null.
     */
    protected String[] _savedGroups = null;
   
    protected ActionListener _findListener = null;
    protected ActionListener _findNextListener = null;
    protected ActionListener _findPreviousListener = null;
    protected ActionListener _replaceAndFindNextListener = null;
    protected ActionListener _replaceAndFindPreviousListener = null;
    protected ActionListener _replaceAllListener = null;         
   
    /** Helps to maintain and notify listeners of status changes. */
    protected PropertyChangeSupport _propertyChangeSupport = null;
   
    protected static final int COMMAND_FIND = 0;    
    protected static final int COMMAND_REPLACE_AND_FIND = 1;
    protected static final int COMMAND_REPLACE_ALL = 2;
    
    /** Search in the forwards direction. */
    protected static final int DIRECTION_FORWARDS = 0;
    
    /** Search in the backwards direction. */
    protected static final int DIRECTION_BACKWARDS = 1;    

    /** Indicates this instance hasn't been used to search yet.
     */                
    public static final int STATUS_INITIAL = 0;
    
    /** Indicates the search text was found. The corresponding extra info
     *  is an int array with two elements: the start and end indices of
     *  the found text, relative to the beginning of the document.
     */
    public static final int STATUS_FOUND = 1;
    
    /** Indicates that the last find operation hit the beginning or end
     *  of the document without finding the search text. The corresponding
     *  extra info is a <code>Boolean</code> which is true iff it was the
     *  beginning that was reached.
     */
    public static final int STATUS_HIT_BOUNDARY = 2;
    
    /** Indicates the search text was not found during the last find operation.
     *  If failure occurred because we wanted to search in a selection, and
     *  there was no selection, the corresponding extra info is
     *  <code>EMPTY_SELECTION</code>. Otherwise, it is <code>null</code>.
     */    
    public static final int STATUS_NOT_FOUND = 3;
    
    /** Indicates an error occured compiling the search text as a regex
     *  pattern. The corresponding extra info is the
     *  <code>PatternSyntaxException</code> that was thrown.
     */
    public static final int STATUS_PATTERN_ERROR = 4;  

    /** Indicates that a replacement operation succeeded. The corresponding
     *  extra info is an int array with two elements: the start and end indices
     *  of the replacement text, relative to the beginning of the document.
     */    
    public static final int STATUS_REPLACED_TEXT = 5;
 
    /** Indicates that a replace all operation succeeded. The corresponding
     *  extra info is an <code>Integer</code> which is the number of 
     *  occurances that were replaced.
     */
    public static final int STATUS_REPLACE_ALL_FINISHED = 6;
 
    /** A value for the extra info that indicates that the last search 
     *  failed because we were searching in a selection, but the selection
     *  was empty.
     */
    public static final Object EMPTY_SELECTION = new Object() {
      public String toString() {
          return "EMPTY_SELECTION";
      }
    };
                      
    private static final Logger _logger = Logger.getLogger(TextSearcher.class);
}
