package com.taco.swinger.text2gui.appbase;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ResourceBundle;
import javax.swing.*;

import javax.swing.JFrame;
import javax.swing.UIManager;
import com.taco.data.INoReturnObservableMap;
import com.taco.util.ChainedResourceBundleFactory;
import com.taco.text.*;
import com.taco.swinger.SwingInvokeProxyFactory;
import com.taco.swinger.text2gui.*;

/** A base class for text2gui applications. This class is inteded to be used
 *  with resource bundles that inherit from AppBaseResourceBundle.
 */
public abstract class AppBase implements PropertyChangeListener {
    public AppBase() { }

    /** Respond to properties in the argument map changing. */    
    public void propertyChange(PropertyChangeEvent event) { }
    
    /** Initialize the GUI. */
    public void initGUI() {
        _setLookAndFeel();
        _makeBundle();
        _makeArgMap();
        _makeFrame();
        _addArgMapListeners();
        _displayFrame();
    }
    
    /** Initialize the GUI in the event dispatch thread. This is intended to be
     *  called by <code>main()</code> with a concrete implementation of this
     *  class.
     */
    public static void start(final AppBase app, final String[] args) {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                app.initGUI();                
            }
        });        
    }            
    
    /** Set the look and feel. This implementation uses the system look and
     *  feel.
     */
    protected void _setLookAndFeel() {
        try {
            UIManager.setLookAndFeel(
              UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            System.err.println("Warning: Can't set look and feel to system");
        }        
    }
    
    /** Create the resource bundle used for creating the GUI. This
     *  implementation appends "ResourceBundle" to the class name and passes
     *  the result to <code>getBundle()</code> of the default instance of
     *  <code>ChainedResourceBundleFactory</code>. This implementation should
     *  suffice for most users.
     */
    protected ResourceBundle _makeBundle() {
        _bundle = ChainedResourceBundleFactory.DEFAULT_INSTANCE.getBundle(
          getClass().getName() + "ResourceBundle");
        return _bundle;  
    }
    
    /** Create the argument map and put the desired values into it. This
     *  implementation creates a map, then puts the this instance as a global
     *  with the name "model". Subclasses should use this map to pass
     *  application-specific parameters to the component converter.
     */    
    protected INoReturnObservableMap _makeArgMap() {
        _argMap = SwingInvokeProxyFactory.makeSwingInvokeProxyMap();
        GlobalUtilities.putGlobal("model", this, _argMap);        
        return _argMap;
    }        
    
    /** Create the main application frame. This implementation first reads the
     *  globals key with an instance collection converter. It then tries to
     *  convert the "mainFrame" resource bundle key. If an exception occurs,
     *  the stack trace is printed and the application exits with a return code
     *  of -1.
     */    
    protected JFrame _makeFrame() {
        try {
            CollectionConverter.INSTANCE_COLLECTION_CONVERTER.toObject(
              _bundle, "globals", _argMap, null);
            _frame = (JFrame) DispatchingComponentConverter.DEFAULT_INSTANCE.
              toObject(_bundle, "mainFrame", _argMap, null);
        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
        return _frame;
    }
        
    /** Add listeners to the argument map, so that
     *  <code>propertyChange()</code> can respond to changes to values mapped
     *  to observed keys.
     */        
    protected void _addArgMapListeners() { }
    
    /** Display the main frame. */ 
    protected void _displayFrame() {
        _frame.pack();
        _frame.setVisible(true);   
    }
    
    /** The main frame of the application. */
    protected JFrame _frame = null;
    
    /** The resource bundle used to create the GUI. */
    protected ResourceBundle _bundle = null;
    
    /** The argument map used to create the GUI and to communicate with it. */
    protected INoReturnObservableMap _argMap = null;        
}
